@@ -13,6 +13,7 @@
 # limitations under the License.
 
 import contextlib, os.path, re, tempfile
+import subprocess
 
 from ..linkers import StaticLinker
 from .. import coredata
@@ -48,7 +49,7 @@ cpp_suffixes = lang_suffixes['cpp'] + ('h',)
 c_suffixes = lang_suffixes['c'] + ('h',)
 # List of languages that can be linked with C code directly by the linker
 # used in build.py:process_compilers() and build.py:get_dynamic_linker()
-clike_langs = ('objcpp', 'objc', 'd', 'cpp', 'c', 'fortran',)
+clike_langs = ('d', 'objcpp', 'cpp', 'objc', 'c', 'fortran',)
 clike_suffixes = ()
 for _l in clike_langs:
     clike_suffixes += lang_suffixes[_l]
@@ -136,8 +137,11 @@ gnulike_buildtype_linker_args = {'plain': [],
 msvc_buildtype_linker_args = {'plain': [],
                               'debug': [],
                               'debugoptimized': [],
-                              'release': [],
-                              'minsize': ['/INCREMENTAL:NO'],
+                              # The otherwise implicit REF and ICF linker
+                              # optimisations are disabled by /DEBUG.
+                              # REF implies ICF.
+                              'release': ['/OPT:REF'],
+                              'minsize': ['/INCREMENTAL:NO', '/OPT:REF'],
                               }
 
 java_buildtype_args = {'plain': [],
@@ -223,9 +227,8 @@ base_options = {'b_pch': coredata.UserBooleanOption('b_pch', 'Use precompiled he
                 'b_colorout': coredata.UserComboOption('b_colorout', 'Use colored output',
                                                        ['auto', 'always', 'never'],
                                                        'always'),
-                'b_ndebug': coredata.UserBooleanOption('b_ndebug',
-                                                       'Disable asserts',
-                                                       False),
+                'b_ndebug': coredata.UserComboOption('b_ndebug', 'Disable asserts',
+                                                     ['true', 'false', 'if-release'], 'false'),
                 'b_staticpic': coredata.UserBooleanOption('b_staticpic',
                                                           'Build static libraries as position independent',
                                                           True),
@@ -252,7 +255,7 @@ vs32_instruction_set_args = {'mmx': ['/arch:SSE'], # There does not seem to be a
                              'avx': ['/arch:AVX'],
                              'avx2': ['/arch:AVX2'],
                              'neon': None,
-}
+                             }
 
 # The 64 bit compiler defaults to /arch:avx.
 vs64_instruction_set_args = {'mmx': ['/arch:AVX'],
@@ -272,7 +275,7 @@ def sanitizer_compile_args(value):
     if value == 'none':
         return []
     args = ['-fsanitize=' + value]
-    if value == 'address':
+    if 'address' in value: # For -fsanitize=address,undefined
         args.append('-fno-omit-frame-pointer')
     return args
 
@@ -312,7 +315,7 @@ def get_base_compile_args(options, compiler):
     except KeyError:
         pass
     try:
-        if options['b_ndebug'].value:
+        if options['b_ndebug'].value == 'true' or (options['b_ndebug'].value == 'if-release' and options['buildtype'] == 'release'):
             args += ['-DNDEBUG']
     except KeyError:
         pass
@@ -450,9 +453,9 @@ class CompilerArgs(list):
         Returns whether the argument can be safely de-duped. This is dependent
         on three things:
 
-        a) Whether an argument can be 'overriden' by a later argument.  For
+        a) Whether an argument can be 'overridden' by a later argument.  For
            example, -DFOO defines FOO and -UFOO undefines FOO. In this case, we
-           can safely remove the previous occurance and add a new one. The same
+           can safely remove the previous occurrence and add a new one. The same
            is true for include paths and library paths with -I and -L. For
            these we return `2`. See `dedup2_prefixes` and `dedup2_args`.
         b) Arguments that once specified cannot be undone, such as `-c` or
@@ -510,10 +513,10 @@ class CompilerArgs(list):
                     continue
                 i = self.index(each)
                 if group_start < 0:
-                    # First occurance of a library
+                    # First occurrence of a library
                     group_start = i
             if group_start >= 0:
-                # Last occurance of a library
+                # Last occurrence of a library
                 self.insert(i + 1, '-Wl,--end-group')
                 self.insert(group_start, '-Wl,--start-group')
         return self.compiler.unix_args_to_native(self)
@@ -547,15 +550,15 @@ class CompilerArgs(list):
             raise TypeError('can only concatenate list (not "{}") to list'.format(args))
         for arg in args:
             # If the argument can be de-duped, do it either by removing the
-            # previous occurance of it and adding a new one, or not adding the
-            # new occurance.
+            # previous occurrence of it and adding a new one, or not adding the
+            # new occurrence.
             dedup = self._can_dedup(arg)
             if dedup == 1:
                 # Argument already exists and adding a new instance is useless
                 if arg in self or arg in pre or arg in post:
                     continue
             if dedup == 2:
-                # Remove all previous occurances of the arg and add it anew
+                # Remove all previous occurrences of the arg and add it anew
                 if arg in self:
                     self.remove(arg)
                 if arg in pre:
@@ -597,20 +600,24 @@ class Compiler:
     # compiler or the C library. Currently only used for MSVC.
     ignore_libs = ()
 
-    def __init__(self, exelist, version):
+    def __init__(self, exelist, version, **kwargs):
         if isinstance(exelist, str):
             self.exelist = [exelist]
         elif isinstance(exelist, list):
             self.exelist = exelist
         else:
             raise TypeError('Unknown argument to Compiler')
-        # In case it's been overriden by a child class already
+        # In case it's been overridden by a child class already
         if not hasattr(self, 'file_suffixes'):
             self.file_suffixes = lang_suffixes[self.language]
         if not hasattr(self, 'can_compile_suffixes'):
             self.can_compile_suffixes = set(self.file_suffixes)
         self.default_suffix = self.file_suffixes[0]
         self.version = version
+        if 'full_version' in kwargs:
+            self.full_version = kwargs['full_version']
+        else:
+            self.full_version = None
         self.base_options = []
 
     def __repr__(self):
@@ -712,6 +719,13 @@ class Compiler:
             'Language {} does not support has_multi_arguments.'.format(
                 self.get_display_language()))
 
+    def get_supported_arguments(self, args, env):
+        supported_args = []
+        for arg in args:
+            if self.has_argument(arg, env):
+                supported_args.append(arg)
+        return supported_args
+
     def get_cross_extra_flags(self, environment, link):
         extra_flags = []
         if self.is_cross and environment:
@@ -809,6 +823,16 @@ class Compiler:
     def get_instruction_set_args(self, instruction_set):
         return None
 
+    def build_osx_rpath_args(self, build_dir, rpath_paths, build_rpath):
+        if not rpath_paths and not build_rpath:
+            return []
+        # On OSX, rpaths must be absolute.
+        abs_rpaths = [os.path.join(build_dir, p) for p in rpath_paths]
+        if build_rpath != '':
+            abs_rpaths.append(build_rpath)
+        args = ['-Wl,-rpath,' + rp for rp in abs_rpaths]
+        return args
+
     def build_unix_rpath_args(self, build_dir, from_dir, rpath_paths, build_rpath, install_rpath):
         if not rpath_paths and not install_rpath and not build_rpath:
             return []
@@ -825,14 +849,24 @@ class Compiler:
         paths = ':'.join([os.path.join('$ORIGIN', p) for p in rel_rpaths])
         # Build_rpath is used as-is (it is usually absolute).
         if build_rpath != '':
-            paths += ':' + build_rpath
+            if paths != '':
+                paths += ':'
+            paths += build_rpath
         if len(paths) < len(install_rpath):
             padding = 'X' * (len(install_rpath) - len(paths))
             if not paths:
                 paths = padding
             else:
                 paths = paths + ':' + padding
-        args = ['-Wl,-rpath,' + paths]
+        args = []
+        if mesonlib.is_dragonflybsd():
+            # This argument instructs the compiler to record the value of
+            # ORIGIN in the .dynamic section of the elf. On Linux this is done
+            # by default, but is not on dragonfly for some reason. Without this
+            # $ORIGIN in the runtime path will be undefined and any binaries
+            # linked against local libraries will fail to resolve them.
+            args.append('-Wl,-z,origin')
+        args.append('-Wl,-rpath,' + paths)
         if get_compiler_is_linuxlike(self):
             # Rpaths to use while linking must be absolute. These are not
             # written to the binary. Needed only with GNU ld:
@@ -840,9 +874,17 @@ class Compiler:
             # Not needed on Windows or other platforms that don't use RPATH
             # https://github.com/mesonbuild/meson/issues/1897
             lpaths = ':'.join([os.path.join(build_dir, p) for p in rpath_paths])
-            args += ['-Wl,-rpath-link,' + lpaths]
+
+            # clang expands '-Wl,rpath-link,' to ['-rpath-link'] instead of ['-rpath-link','']
+            # This eats the next argument, which happens to be 'ldstdc++', causing link failures.
+            # We can dodge this problem by not adding any rpath_paths if the argument is empty.
+            if lpaths.strip() != '':
+                args += ['-Wl,-rpath-link,' + lpaths]
         return args
 
+    def thread_flags(self, env):
+        return []
+
 
 GCC_STANDARD = 0
 GCC_OSX = 1
@@ -869,7 +911,11 @@ def get_gcc_soname_args(gcc_type, prefix, shlib_name, suffix, path, soversion, i
     elif gcc_type == GCC_OSX:
         if is_shared_module:
             return []
-        return ['-install_name', os.path.join(path, 'lib' + shlib_name + '.dylib')]
+        install_name = prefix + shlib_name
+        if soversion is not None:
+            install_name += '.' + soversion
+        install_name += '.dylib'
+        return ['-install_name', os.path.join('@rpath', install_name)]
     else:
         raise RuntimeError('Not implemented yet.')
 
@@ -910,6 +956,41 @@ def get_largefile_args(compiler):
         # those features explicitly.
     return []
 
+# TODO: The result from calling compiler should be cached. So that calling this
+# function multiple times don't add latency.
+def gnulike_default_include_dirs(compiler, lang):
+    if lang == 'cpp':
+        lang = 'c++'
+    env = os.environ.copy()
+    env["LC_ALL"] = 'C'
+    cmd = compiler + ['-x{}'.format(lang), '-E', '-v', '-']
+    p = subprocess.Popen(
+        cmd,
+        stdin=subprocess.DEVNULL,
+        stderr=subprocess.PIPE,
+        stdout=subprocess.PIPE,
+        env=env
+    )
+    stderr = p.stderr.read().decode('utf-8')
+    parse_state = 0
+    paths = []
+    for line in stderr.split('\n'):
+        if parse_state == 0:
+            if line == '#include "..." search starts here:':
+                parse_state = 1
+        elif parse_state == 1:
+            if line == '#include <...> search starts here:':
+                parse_state = 2
+            else:
+                paths.append(line[1:])
+        elif parse_state == 2:
+            if line == 'End of search list.':
+                break
+            else:
+                paths.append(line[1:])
+    if len(paths) == 0:
+        mlog.warning('No include directory found parsing "{cmd}" output'.format(cmd=" ".join(cmd)))
+    return paths
 
 class GnuCompiler:
     # Functionality that is common to all GNU family compilers.
@@ -962,7 +1043,7 @@ class GnuCompiler:
         return 'gch'
 
     def split_shlib_to_parts(self, fname):
-        return os.path.split(fname)[0], fname
+        return os.path.dirname(fname), fname
 
     def get_soname_args(self, prefix, shlib_name, suffix, path, soversion, is_shared_module):
         return get_gcc_soname_args(self.gcc_type, prefix, shlib_name, suffix, path, soversion, is_shared_module)
@@ -991,6 +1072,9 @@ class GnuCompiler:
     def get_instruction_set_args(self, instruction_set):
         return gnulike_instruction_set_args.get(instruction_set, None)
 
+    def get_default_include_dirs(self):
+        return gnulike_default_include_dirs(self.exelist, self.language)
+
 
 class ClangCompiler:
     def __init__(self, clang_type):
@@ -1075,6 +1159,9 @@ class ClangCompiler:
     def get_instruction_set_args(self, instruction_set):
         return gnulike_instruction_set_args.get(instruction_set, None)
 
+    def get_default_include_dirs(self):
+        return gnulike_default_include_dirs(self.exelist, self.language)
+
 
 # Tested on linux for ICC 14.0.3, 15.0.6, 16.0.4, 17.0.1
 class IntelCompiler:
@@ -1104,10 +1191,10 @@ class IntelCompiler:
                 self.lang_header, '-include', header, '-x', 'none']
 
     def get_pch_name(self, header_name):
-        return os.path.split(header_name)[-1] + '.' + self.get_pch_suffix()
+        return os.path.basename(header_name) + '.' + self.get_pch_suffix()
 
     def split_shlib_to_parts(self, fname):
-        return os.path.split(fname)[0], fname
+        return os.path.dirname(fname), fname
 
     def get_soname_args(self, prefix, shlib_name, suffix, path, soversion, is_shared_module):
         if self.icc_type == ICC_STANDARD:
@@ -1125,3 +1212,6 @@ class IntelCompiler:
         # if self.icc_type == ICC_OSX:
         #     return ['-bundle']
         return ['-shared']
+
+    def get_default_include_dirs(self):
+        return gnulike_default_include_dirs(self.exelist, self.language)
